{
Autor: Rafael Alberto Moreno Parra
Sitio Web: http://darwin.50webs.com
Tema: Regresin Simblica
Licencia: GPL

Objetivo: Dada una serie de datos X, Y encontrar la mejor curva de tendencia
}
unit RegresionA;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, Poblacion, Grids, ComCtrls;

type
  TGenetico = class(TThread)

    //Almacena el ndice de registro del mejor individuo encontrado
    MejorIndividuo: integer;

    //Activa y detiene el hilo de ejecucin
    proceso: bool;

    //Objeto que sostiene la poblacin de individuos
    objPoblacion: TPoblacion;

    //Contador de ciclos para refrescar la pantalla (grfico lineal)
    Progreso: integer;

    //Componentes grficos que se refrescan peridicamente durante el proceso

    //La mejor ecuacin encontrada durante el proceso
    txtEcuacion: TEdit;

    //Aproximacin
    txtAproximacion: TEdit;

    //Hoja de clculo que muestra el histrico de las ecuaciones encontradas
    griProgreso: TStringGrid;

    //Muestra el progreso del proceso
    lblProgreso: TLabel;

    //Proceso que ejecuta en un hilo
    procedure Execute; override;

    //Cada cierto perodo la pantalla se refresca con los nuevos resultados
    procedure ActualizarPantalla;
  end;

  TfrmRegresion = class(TForm)
    btnProceso: TButton;
    btnCargaLista: TButton;
    lblEcuacion: TLabel;
    lblAproximacion: TLabel;
    txtEcuacion: TEdit;
    txtAproximacion: TEdit;
    lblAviso: TLabel;
    pgcDatos: TPageControl;
    tabDatosLeidos: TTabSheet;
    griAproxima: TStringGrid;
    btnAcercade: TButton;
    btnDetener: TButton;
    lblProgreso: TLabel;
    tabProgreso: TTabSheet;
    griProgreso: TStringGrid;
    lblPoblacion: TLabel;
    cboTamPoblacion: TComboBox;
    procedure btnCargaListaClick(Sender: TObject);
    procedure btnProcesoClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure griAproximaDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure btnAcercadeClick(Sender: TObject);
    procedure btnDetenerClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    { Private declarations }
    Genetico: TGenetico;
  public
    { Public declarations }
  end;

var
  frmRegresion: TfrmRegresion;

implementation

{$R *.dfm}

procedure TfrmRegresion.btnCargaListaClick(Sender: TObject);
var
  dialogoArchivo: TOpenDialog;
  cont: integer;
  valorX, valorY, valorZ: double;
begin
  //Crea el objeto de ventana de dilogo para abrir archivos
  dialogoArchivo := TOpenDialog.Create(self);

  //La ventana de dilogo apunta al directorio actual
  dialogoArchivo.InitialDir := GetCurrentDir;

  //Slo archivos existentes pueden operar
  dialogoArchivo.Options := [ofFileMustExist];

  //Filtra los archivos de trios de datos
  dialogoArchivo.Filter := 'Trios de valores|*.trio';

  //Muestra la ventana de dilogo
  if dialogoArchivo.Execute then
  begin
    //Crea el algoritmo gentico
    if Assigned(Genetico) then
    begin
      Freeandnil(Genetico);
    end;
    Genetico := TGenetico.Create(True);
    Genetico.Progreso := 1;
    Genetico.objPoblacion := TPoblacion.Create;
    Genetico.objPoblacion.leeDatosTrios(dialogoArchivo.FileName);
    txtEcuacion.Text := '';
    txtAproximacion.Text := '';
    lblProgreso.Caption := 'Progreso:';

    //Inicializa los controles gridbox
    for cont := 1 to griAproxima.RowCount do
    begin
      griAproxima.Cells[0, cont] := '';
      griAproxima.Cells[1, cont] := '';
      griAproxima.Cells[2, cont] := '';
    end;

    for cont := 1 to griAproxima.RowCount do
    begin
      griProgreso.Cells[0, cont] := '';
      griProgreso.Cells[1, cont] := '';
      griProgreso.Cells[2, cont] := '';
    end;

    griAproxima.RowCount := 2;
    griProgreso.RowCount := 2;
    griAproxima.RowCount := Genetico.objPoblacion.limValores + 1;

    for cont := 1 to Genetico.objPoblacion.limValores do
    begin
      //Lo agrega al grfico
      valorX := Genetico.objPoblacion.valorX[cont];
      valorY := Genetico.objPoblacion.valorY[cont];
	    valorZ := Genetico.objPoblacion.valorZ[cont];

      //Muestra los datos en el componente GridBox
      with griAproxima do
      begin
          //Datos
          Cells[0, cont] := floattostr(valorX);
          Cells[1, cont] := floattostr(valorY);
		      Cells[2, cont] := floattostr(valorZ);
      end;
    end;
    btnProceso.Enabled := true;
    Genetico.lblProgreso := lblProgreso;
    Genetico.griProgreso := griProgreso;
    Genetico.txtEcuacion := txtEcuacion;
    Genetico.txtAproximacion := txtAproximacion;
  end
  else
    ShowMessage('Cancelado');

   //Libera la ventana de dilogo
   dialogoArchivo.Free;
 end;


procedure TfrmRegresion.btnProcesoClick(Sender: TObject);
var
  tamPoblacion: integer;
begin
    //Leer los parmetros de simulacin
    tamPoblacion := strtoint(cboTamPoblacion.Text);

    //Crea la poblacin inicial
    Genetico.objPoblacion.ConfProceso(tamPoblacion);
    Genetico.MejorIndividuo := -1;

    //Activa los botones de detener y continuar proceso
    btnDetener.Enabled := true;
    btnCargaLista.Enabled := false;
    btnProceso.Enabled := false;

    //Inicia el hilo
    Genetico.Resume;
end;

//Este mtodo es ejecutado por el hilo
procedure TGenetico.Execute;
begin
  inherited;
  proceso := true;
  while proceso do
  begin
    objPoblacion.Proceso();
    if progreso mod 1000 = 0 then Synchronize(ActualizarPantalla);
    Inc(progreso);
  end;
end;

//Este procedimiento es ejecutado cada perodo para mostrar los nuevos resultados
procedure TGenetico.ActualizarPantalla;
begin
  lblProgreso.Caption := 'Ciclo: ' + inttostr(progreso);

  //Si hay un mejor individuo, lo presenta
  if objPoblacion.mejorIndividuo = self.MejorIndividuo then Exit;
  self.MejorIndividuo := objPoblacion.mejorIndividuo;
  txtEcuacion.Text := objPoblacion.RetornaExpresion(self.MejorIndividuo);
  txtAproximacion.Text := floattostr(objPoblacion.CalculaAproximacion(self.MejorIndividuo));

  //Muestra el nuevo individuo que es mejor
  with griProgreso do
  begin
    Cells[0, griProgreso.RowCount-1] := inttostr(progreso);
    Cells[1, griProgreso.RowCount-1] := txtEcuacion.Text;
    Cells[2, griProgreso.RowCount-1] := txtAproximacion.Text;
  end;
  griProgreso.RowCount := griProgreso.RowCount + 1;
end;

procedure TfrmRegresion.btnDetenerClick(Sender: TObject);
begin
    //Detiene el hilo
    Genetico.proceso := false;
    Genetico.WaitFor;

    //Activa el botn de continuar con el proceso y cargar lista
    btnCargaLista.Enabled := true;
    btnDetener.Enabled := false;
end;

//Inicializa el Grid
procedure TfrmRegresion.FormCreate(Sender: TObject);
var
  cont: integer;
begin
  //Inicializa el Grid
  with griAproxima do
  begin
    //Etiqueta de las columnas
    Cells[0, 0] := 'Valor X';
    Cells[1, 0] := 'Valor Y';
    Cells[2, 0] := 'Valor Z';
  end;

  //Inicializa el Grid
  griProgreso.ColWidths[0] := 70;
  griProgreso.ColWidths[1] := 700;
  with griProgreso do
  begin
    //Etiqueta de las columnas
    Cells[0, 0] := 'Ciclo';
    Cells[1, 0] := 'Mejor ecuacin';
    Cells[2, 0] := 'Aproximacin';
  end;

  //Llena el combobox de las opciones de tamao de poblacin
  cont := 1000;
  repeat
    cboTamPoblacion.Items.Add(inttostr(cont));
    cont := cont + 1000;
  until cont > 10000;
end;

//Justifica a la derecha los valores numricos en el Grid
procedure TfrmRegresion.griAproximaDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
var
  text: string;
  WidthOfText: integer;
  WidthOfCell: integer;
  LeftOffset: integer;
begin
 if ARow > 0 then
   with griAproxima do begin
    text := Cells[ACol,ARow];
    WidthOfText := Canvas.TextWidth(text);
    WidthOfCell := ColWidths[ACol];
    LeftOffset := WidthOfCell - WidthOfText;
    Canvas.TextRect(Rect,Rect.Left+LeftOffset,Rect.Top,text);
   end;
end;

procedure TfrmRegresion.btnAcercadeClick(Sender: TObject);
begin
  ShowMessage('Autor: Rafael Alberto Moreno Parra' + #10#13 + 'Pgina web: http://darwin.50webs.com' + #10#13 + 'Investigacin: Deteccin de patrones en series de datos: Regresin Simblica' + #10#13 + 'Versin: 1.5 (Agosto de 2013)' + #10#13 + 'Modo de uso: Lee los datos de un archivo plano (*.trio) y busca la ecuacin z=f(x,y) que ms se aproxime a esa serie' + #10#13 + 'Licencia: GPL');
end;


procedure TfrmRegresion.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if Genetico <> nil then
  begin
    Genetico.proceso := false;
    Genetico.Free;
  end;
end;

end.
